unit uthreadworld;

{$mode objfpc}{$H+}

interface

uses  // -dUseCThreads
  LCLIntf, Classes, Sysutils, Dialogs
  {, Forms} // Threads should not use unit forms !
  ;

Const

  wrSignaled = 0;
  wrTimeout  = 1;
  wrAbandoned= 2;
  wrError    = 3;

type
  TWorld = class;

  { TLifeThread }

  // Simple thread model for parallel running
  TLifeThread = class(TThread)
  private
    FWorld: TWorld;
    FName: string;
    FStatus: string;
    _fireEvent:pEventState;
    _killEvent:pEventState;
    procedure _status;
  public
    Value:integer;
    constructor create(owner:TWorld); overload;
    destructor destroy; override;
    procedure Execute; override;
    property Name:string read FName write FName;
  end;

  // Simple parallel threads

  { TWorld }

  TonStatusEvent = procedure(Sender:TObject; aStatus:string) of Object;

  TWorld = class(TComponent)
  private
    FList:TThreadList;
    FonStatus: TonStatusEvent;
    FStatus:string;
    FView:TStringList;
    FLock1:boolean;
    FLock2:boolean;
    FResources:integer;
    FdontStarve:boolean;
    function getThreadCount: integer;
    function getThreadIndex(aThread: TLifeThread): integer;
    function getLocked: boolean;
    function getThreadsRunning: integer;
    function getView:string;
    function getResources:integer;
  public
    constructor create(aOwner:TComponent); overload;
    destructor destroy; override;
    procedure fire;
    procedure terminateAllLifes;
    procedure newLife;
    property View:string read getView;
    property ThreadsRunning:integer read getThreadsRunning;
    property Locked:boolean read getLocked;
    property Resources:integer read getResources write FResources;
    property onStatus:TonStatusEvent read FonStatus write FonStatus;
  end;


    {$IFDEF USEWINDOWS}
    var GlobalLock: TRTLCriticalSection; // {windows}
    {$ELSE}
     {$IFDEF WIN64}
       var GlobalLock: QWORD {Win64};  // TCriticalSection = QWORD for WIN64; ????
     {$ELSE}
       var GlobalLock: LongWord {Win32, Linux, ...};
     {$ENDIF}
    {$ENDIF}

implementation

{ TLifeThread }

constructor TLifeThread.create(owner:TWorld);
begin
  inherited create(true {suspended});
  _fireEvent := BasicEventCreate(nil,false,false,'');
  _killEvent := BasicEventCreate(nil,false,false,'');
  // InitializeCriticalSection(CritSect);
end;

destructor TLifeThread.destroy;
begin
  basiceventdestroy(_fireEvent);
  basiceventdestroy(_killEvent);
  // DeleteCriticalSection(CritSect);
  inherited;
end;

procedure TLifeThread._status;
begin
  if assigned(FWorld.FonStatus) then
     FWorld.FonStatus(self, FStatus);
end;

procedure TLifeThread.Execute;
var starved:boolean; actLifeIndex:integer;
begin
  starved := false;
  while not Terminated do
     begin
     if basiceventWaitFor(1, _killEvent) = wrSignaled then
        begin
        basiceventResetEvent(_killEvent);
        FreeOnTerminate := false;
        terminate;
        break;
        end
     else
     if basiceventWaitFor(1, _fireEvent) = wrSignaled then
        begin
        basiceventResetEvent(_fireEvent);
        // add value to common resources
        EnterCriticalSection(GlobalLock);
        with FWorld do
           FResources := FResources + Value;
        LeaveCriticalSection(GlobalLock);
        Value := 0;
        // send status to GUI:
        FStatus := FName + ': reset to zero!';
        synchronize(@_status);
        end
     else
        begin
        Value := Value + 1;
        // receive one unit from common resources
        EnterCriticalSection(GlobalLock);
        with FWorld do
           FResources := FResources -1;
        LeaveCriticalSection(GlobalLock);
        // if all resources are wasted life starves
        if (FWorld.FResources = 0) and not FWorld.FdontStarve then
           begin                           // FdontStarve:=true on ...
           starved := true;                // ... World.terminateAllLifes
           with FWorld do
              begin
              EnterCriticalSection(GlobalLock);
              FResources := FResources + Value; // give back collected resources
              LeaveCriticalSection(GlobalLock);
              //    actLifeIndex := getThreadIndex(self);
              FList.remove(self);
              end;
           end;
        // terminate live if starved
        if starved then
           begin
           freeOnTerminate := true;
           terminate;
           break;
           end;
        end;

     sleep(998);
     end;
end;

{ TWorld }

constructor TWorld.create(aOwner:TComponent);
begin
  FList := TThreadList.create;
  FView := TStringList.create;
  FResources := 1000;
  InitializeCriticalSection(GlobalLock);
  inherited create(aOwner);
end;

destructor TWorld.destroy;
begin
  FList.free;
  FView.free;
  DeleteCriticalSection(GlobalLock);
  inherited;
end;

// not used but perhaps usefully
function TWorld.getThreadIndex(aThread: TLifeThread): integer;
var list: TList;
begin
  list:= FList.Locklist;
  try
    Result := list.indexOf(aThread);
  finally
    FList.Unlocklist;
  end;
end;

// not used but perhaps usefully
function TWorld.getThreadCount: integer;
var list: TList;
begin
  list:= FList.Locklist;
  try
    Result := list.count;
  finally
    FList.Unlocklist;
  end;
end;

function TWorld.getView:string;
var i:integer; list:TList;
begin
  if FLock1 then   // prevent overload by gui
     exit;
  FLock1 := true;
  try
    FView.clear;
    list := FList.Locklist;
    For i := 0 to list.count - 1 do
       FView.add(floatToStr(TLifeThread(list[i]).Value));
    Result := FView.text;
  finally
    FList.UnlockList;
    FLock1 := false;
  end;
end;

function TWorld.getResources: integer;
var i:integer; list:TList;
begin
  if FLock2 then  // prevent overload by gui
     exit;
  FLock2 := true;
  try
    Result := FResources;
  finally
    FLock2 := false;
  end;
end;

// gui should test on 'Locked' before an gui action
function TWorld.getLocked: boolean;
begin
  Result := FLock1 or FLock2 {or FLock3 or ... };
end;

function TWorld.getThreadsRunning: integer;
var list: TList;
begin
  Result := -1;
  list:= FList.Locklist;
  try
    Result := list.count;
  finally
    FList.Unlocklist;
  end;
end;

procedure TWorld.newLife;
var aNewLife:TLifeThread; err_cre:boolean; ix:integer;
begin
  FdontStarve := false;
  err_cre := false;
  try
    aNewLife := TLifeThread.Create(self);
    // This code will assure that any exception which occurred
    // during thread creation will be raised in the WorldThread:
    if Assigned(aNewLife.FatalException) then
       raise aNewLife.FatalException;
  except
    err_cre := true;
    showMessage('Error: can''t create so many threads');
    aNewLife := nil;
  end;
  if err_cre then
     exit;
 // Does not work together with 'waitFor':   <----- !!!!!
 // aNewLife.onTerminate := @lifeTerminated;
 aNewLife.FWorld := self;
 aNewLife.FName := 'life ' + intToStr(random(10000));
 FList.add(aNewLife);
 aNewLife.resume;
end;

procedure TWorld.fire;
var ix:integer; list:TList;
begin
  list := FList.Locklist;
  if list.count > 0 then
     try
       ix := random(list.count);
       basiceventSetEvent(TLifeThread(List[ix])._fireEvent);
     finally
       FList.Unlocklist
     end;
end;

procedure TWorld.terminateAllLifes;
var i, co:integer; aLife:TLifeThread; res:dword; list:TList;
begin
  FdontStarve := true;
  list := FList.locklist;
  try
    for i := list.count - 1 downto 0 do
       begin
       aLife := TLifeThread(list[i]);
       aLife.freeOnTerminate := false;
       // terminate by 'event' (other possibility: 'terminate'):
       basiceventSetEvent(TLifeThread(list[i])._killEvent);
       aLife.waitFor; // don't use 'onTerminate' !!!!
       // res := killthread(aLife.handle);// faster as 'waitFor' but qick&dirty!
       freeAndNil(aLife);
       list.delete(i);
       end;
  finally
    FList.Unlocklist;
  end;
  FStatus := '';
end;

end.
